#include "CyanLightControl.hpp"

#include <qthing>
#include <qthing/mqtt_common.hpp>

#include <string>
#include <cstdio>
#include <cstdlib>
#include <algorithm>
#include <functional>

#include "esp_err.h"
#include "esp_log.h"
#include "driver/ledc.h"


using namespace qthing;


const char *TAG = "CLC";
const std::string delimiter = ":";


static float bytes2float(const char *bytes) {
  float f;

  memcpy(&f, bytes, sizeof(f));

  return f;
}


CyanLight::Controller::Controller(uint16_t baseChannel, uint8_t channelsConfigured) : baseChannel{baseChannel} {

  this->channelsConfigured = std::min(channelsConfigured, CyanLight::MAX_CHANNELS);

  this->channels = new CyanLight::PWMChannel*[this->channelsConfigured];


  this->timer_cfg.duty_resolution = (ledc_timer_bit_t)this->resolution;
  this->timer_cfg.freq_hz = this->frequency;
  this->timer_cfg.speed_mode = LEDC_HIGH_SPEED_MODE;
  this->timer_cfg.timer_num = LEDC_TIMER_0;
  // Required for newer esp-idf:
  //ledc_timer.clk_cfg = LEDC_AUTO_CLK,             // Auto select the source clock
  ledc_timer_config(&this->timer_cfg);

  ledc_fade_func_install(0);


  for (uint8_t i = 0; i < this->channelsConfigured; i++) {
    this->channels[i] = new CyanLight::PWMChannel(i, CyanLight::channelGPIOS[i]);

    char topic[32];
    snprintf(topic, sizeof(topic), "pwm/$%i", i);

    add_message_callback(this->genDeviceTopic(topic), [&, i](std::string message) {
      this->setPWM(i, strtof(message.c_str(), NULL));
      this->callPacketCallback();
    });
  }


  std::string header("fxCyanF");
  uint8_t headerLength = header.length();

  qthing::addUDPPacketCallback(header, [&, headerLength](udpPacket packet) {
    this->handleUnicast(packet.payload + headerLength, packet.length - headerLength);
  });


  std::function<void(std::string)> setCh = [&](std::string message) {
    long int ch = strtol(message.c_str(), NULL, 0);
    this->setBaseChannel(ch);
  };

  std::function<void(std::string)> setFrq = [&](std::string message) {
    long int frq = strtol(message.c_str(), NULL, 0);
    this->setFrequency(frq);
  };

  std::function<void(std::string)> setRes = [&](std::string message) {
    long int res = strtol(message.c_str(), NULL, 0);
    this->setResolution(res);
  };

  std::function<void(std::string)> setFrqRes = [&](std::string message) {
    std::string::size_type found = message.find(delimiter);

    if (found == std::string::npos) {
      ESP_LOGW("CLC", "Invalid message format[ '%s' ]", message.c_str());
      return;
    }

    std::string _frq = message.substr(0, found);
    std::string _res = message.substr(found + delimiter.length());

    long int frq = strtol(_frq.c_str(), NULL, 0);
    long int res = strtol(_res.c_str(), NULL, 0);

    if (frq == 0 || res == 0) {
      ESP_LOGW("CLC", "Invalid message format[ '%s' ]", message.c_str());
      return;
    }

    this->setFrqRes(frq, res);
  };


  std::function<void(std::string)> getBCh = [&](std::string ignored) {
    this->publishBaseChannel();
  };

  std::function<void(std::string)> getChs = [&](std::string ignored) {
    this->publishChannelCount();
  };

  std::function<void(std::string)> getFrq = [&](std::string ignored) {
    this->publishFrequency();
  };

  std::function<void(std::string)> getRes = [&](std::string ignored) {
    this->publishResolution();
  };

  std::function<void(std::string)> getFrqRes = [&](std::string ignored) {
    this->publishFrqRes();
  };


  // device-local setters
  add_message_callback(this->genDeviceTopic("frqres/set"), setFrqRes);
  add_message_callback(this->genDeviceTopic("channel/set"), setCh);
  add_message_callback(this->genDeviceTopic("frequency/set"), setFrq);
  add_message_callback(this->genDeviceTopic("resolution/set"), setRes);

  // device-lokal getters
  add_message_callback(this->genDeviceTopic("frequency/get"), getFrq);
  add_message_callback(this->genDeviceTopic("resolution/get"), getRes);
  add_message_callback(this->genDeviceTopic("frqres/get"), getFrqRes);
  add_message_callback(this->genDeviceTopic("channel/get"), getBCh);
  add_message_callback(this->genDeviceTopic("channelCnt/get"), getChs);

  // global setters
  add_message_callback(this->genServiceTopic("frqres/set"), setFrqRes);
  add_message_callback(this->genServiceTopic("channel/set"), setCh);
  add_message_callback(this->genServiceTopic("frequency/set"), setFrq);
  add_message_callback(this->genServiceTopic("resolution/set"), setRes);

  // global getters
  add_message_callback(this->genServiceTopic("frequency/get"), getFrq);
  add_message_callback(this->genServiceTopic("resolution/get"), getRes);
  add_message_callback(this->genServiceTopic("frqres/get"), getFrqRes);
  add_message_callback(this->genServiceTopic("channel/get"), getBCh);
  add_message_callback(this->genServiceTopic("channelCnt/get"), getChs);
}


bool CyanLight::Controller::handleUnicast(const char *data, std::size_t length) {
  std::size_t size = this->getChannelCount() * sizeof(float);

  int32_t diff = length - size;
  if (diff < 0) {
    ESP_LOGE(TAG, "Invalid data length[ %i ]: Received ΔB = %i bytes too few", length, -diff);
    return false;
  }

  for (uint8_t ch = 0; ch < this->getChannelCount(); ch++) {
    float f = bytes2float(data + ch * sizeof(float));
    this->setPWM(ch, f);
  }

  this->callPacketCallback();

  return true;
}

bool CyanLight::Controller::handleBroadcast(const char *data, std::size_t length) {
  std::size_t size = this->getChannelCount() * sizeof(float);
  std::size_t offset = this->getBaseChannel() * sizeof(float);

  int32_t diff = length - offset - size;
  if (diff < 0) {  // TODO: test thoroughly!
    ESP_LOGE(TAG, "Invalid data length[ %i ]: Received ΔB = %i bytes too few", length, -diff);
    return false;
  }

  for (uint8_t ch = 0; ch < this->getChannelCount(); ch++) {
    float f = bytes2float(data + offset + ch * sizeof(float));
    this->setPWM(ch, f);
  }

  this->callPacketCallback();

  return true;
}


void CyanLight::Controller::setPWM(uint8_t channel, float value) {

  if (channel >= this->channelsConfigured) {
    ESP_LOGW(TAG, "ChannelID out of range: channel[ %i ]", channel);
    return;
  }

  value = this->gammaCorrector(value);

  if (value < 0.0f) value = 0.0f;
  if (value > 1.0f) value = 1.0f;

  uint32_t maxPWM = 1 << (this->resolution - 1);
  uint32_t pwm = (uint32_t)( value * (maxPWM - 1) );

  this->channels[channel]->setPWM(pwm);
}


void CyanLight::Controller::setGammaCorrector(GammaCorrector gammaCorrector) {
  this->gammaCorrector = gammaCorrector;
}


bool CyanLight::Controller::setFrqRes(uint32_t frq_hz, uint8_t res_bits) {

  this->timer_cfg.duty_resolution = (ledc_timer_bit_t)res_bits;
  this->timer_cfg.freq_hz = frq_hz;

  if (ledc_timer_config(&this->timer_cfg) == ESP_OK) {
    this->frequency = frq_hz;
    this->resolution = res_bits;

    ESP_LOGI(TAG, "frequency[ %i Hz ]  resolution[ %i bit ]", frq_hz, res_bits);

    return true;

  } else {

    ESP_LOGW(TAG, "Invalid frequency[ %i Hz ] and resolution[ %i bit ] setting!", frq_hz, res_bits);
    return false;
  }
}


void CyanLight::Controller::setBaseChannel(uint16_t baseChannel) {
  this->baseChannel = baseChannel;
}

uint16_t CyanLight::Controller::getBaseChannel() {
  return this->baseChannel;
}

uint8_t CyanLight::Controller::getChannelCount() {
  return this->channelsConfigured;
}

bool CyanLight::Controller::setFrequency(uint32_t frq_hz) {
  return this->setFrqRes(frq_hz, this->resolution);
}

uint32_t CyanLight::Controller::getFrequency() {
  return this->frequency;
}

bool CyanLight::Controller::setResolution(uint8_t res_bits) {
  return this->setFrqRes(this->frequency, res_bits);
}

uint8_t CyanLight::Controller::getResolution() {
  return this->resolution;
}

void CyanLight::Controller::setPacketHandledCallback(PacketHandledCallback callback) {
  this->packetCallback = callback;
}


std::string CyanLight::Controller::genDeviceTopic(const char *suffix) {
  return std::string(DEVICE_NAMESPACE + "fxCyanF/") + std::string(suffix);
}

std::string CyanLight::Controller::genServiceTopic(const char *suffix) {
  return std::string("service/fxCyanF/") + std::string(suffix);
}


void CyanLight::Controller::callPacketCallback() {
  this->packetCallback();
}


void CyanLight::Controller::publishBaseChannel() {
  char tmp[16];
  snprintf(tmp, sizeof(tmp), "%i", this->getBaseChannel());
  publish_message(DEVICE_NAMESPACE + "CyanLight/channel", tmp);
}

void CyanLight::Controller::publishChannelCount() {
  char tmp[16];
  snprintf(tmp, sizeof(tmp), "%i", this->getChannelCount());
  publish_message(DEVICE_NAMESPACE + "CyanLight/channelCnt", tmp);
}

void CyanLight::Controller::publishFrequency() {
  char tmp[16];
  snprintf(tmp, sizeof(tmp), "%i", this->getFrequency());
  publish_message(DEVICE_NAMESPACE + "CyanLight/frequency", tmp);
}

void CyanLight::Controller::publishResolution() {
  char tmp[16];
  snprintf(tmp, sizeof(tmp), "%i", this->getResolution());
  publish_message(DEVICE_NAMESPACE + "CyanLight/resolution", tmp);
}

void CyanLight::Controller::publishFrqRes() {
  char tmp[32];
  snprintf(tmp, sizeof(tmp), "%i%s%i", this->getFrequency(), delimiter.c_str(), this->getResolution());
  publish_message(DEVICE_NAMESPACE + "CyanLight/frqres", tmp);
}

